unit FILTER; {Unit 9 of SEASIDE.PAS}

interface

  uses
    GLOBALS, UTILS, MATHS, LEGENDS, DATASEG, SCREEN, POST;

  procedure Convolution (var alpha, beta: seasVector;
                  p, k: integer);

  procedure Laurent (var mu, gamma: seasVector;
                  q: integer);

  procedure Minit (var mu: vector;
                  var varEpsilon: real;
                  gamma: vector;
                  q: integer);

  procedure SeasonalPoles (var s: integer;
                  var rho: real;
                  var poles: complexVector);

  procedure SeasonalZeros (var s: integer;
                  var zeros: complexVector);

  procedure SprimeY (var y: longVector;
                  s, Tcap: integer);

  procedure SprimeY2 (q, n: integer;
                  var alpha: seasVector;
                  var y: longVector);

  procedure Sy2 (q, n: integer;
                  var alpha: seasVector;
                  var y: longVector);

  procedure ApplySingleSeasonalFilter (var s, Tcap: integer;
                  var phi, sigmaVec: seasVector;
                  var seasonVec, adjustedVec: longVector);

  procedure ApplySmoothingFilter (var s, q, Tcap: integer;
                  var phi, phi2, sigmaVec, muVec: seasVector;
                  var seasonVec, adjustedVec: longVector);

  procedure ApplyDoubleSeasonalFilter (var s, Tcap: integer;
                  var phi1, phi2, sigmaVec1, sigmaVec2: seasVector;
                  var seasonVec, adjustedVec: longVector);

  procedure ApplyTripleSeasonalFilter (var s, Tcap: integer;
                  var phi, phi1, phi2, sigmaVec, sigmaVec1, sigmaVec2: seasVector;
                  var seasonVec, adjustedVec: longVector);

  procedure TrendPlusSeason (var polyParas: polyParameters;
                  var dataBOX, adjustedBOX: rBOX;
                  var dataVector, adjustedVec: longVector;
                  var SCREENBOX, POSTBOX: IBOX);

implementation

{_________________________________________________}

  procedure Convolution (var alpha, beta: seasVector;
                  p, k: integer);
    var
      gamma: real;
      i, j, r, s: integer;

  begin

    for j := p + k downto 0 do
      begin {j}
        s := Min(j, p);
        r := Max(0, j - k);
        gamma := 0.0;
        for i := r to s do
          gamma := gamma + alpha[i] * beta[j - i];
        beta[j] := gamma;
      end; {j}

  end; {Convolution}

{_________________________________________________}

  procedure Laurent (var mu, gamma: seasVector;
                  q: integer);

    var
      i, j: integer;

  begin
    for i := 0 to q do
      begin {i}
        gamma[i] := 0.0;
        for j := 0 to q - i do
          gamma[i] := gamma[i] + mu[j] * mu[j + i];
      end; {i}
  end; {Laurent}

{________________________________________________}

  procedure Minit (var mu: vector;
                  var varEpsilon: real;
                  gamma: vector;
                  q: integer);

    var
      D: matrix;
      delta, f: vector;
      i, j, iterations: integer;
      convergence: boolean;

  {-------------------------------------------}

    function CheckDelta (q: integer;
                    var delta, mu: vector): boolean;

      var
        i: integer;
        muNorm, deltaNorm: real;

    begin
      muNorm := 0.0;
      deltaNorm := 0.0;
      for i := 0 to q do
        begin
          muNorm := muNorm + sqr(mu[i]);
          deltaNorm := deltaNorm + Sqr(delta[i])
        end;
      if (deltaNorm / muNorm) > 1.0E-20 then
        CheckDelta := false
      else
        CheckDelta := true;

    end; {Checkdelta}

  {------------------------------------------}

  begin {PROCEDURE Minit}

  {initialise the vector mu }
    for i := 0 to q do
      mu[i] := 0.0;
    mu[0] := Sqrt(gamma[0]);
    convergence := false;
    iterations := 0;

    while (convergence = false) and (iterations < 25) do
      begin

  {Form the matrix of derivatives}
        for i := 0 to q do
          for j := 0 to q do
            begin
              D[i, j] := 0.0;
              if (j - i) >= 0 then
                D[i, j] := mu[j - i];
              if (i + j) <= q then
                D[i, j] := D[i, j] + mu[i + j];
            end;

  {Find the function value}
        for j := 0 to q do
          begin
            f[j] := gamma[j];
            for i := 0 to q - j do
              f[j] := f[j] - mu[i] * mu[i + j]
          end;

  {Find the updating vector}
        LUsolve(0, q + 1, D, delta, f);

  {Update the value of mu}
        for i := 0 to q do
          mu[i] := mu[i] + delta[i];
        iterations := iterations + 1;

  {Check for convergence}
        convergence := CheckDelta(q, delta, mu);
      end; {while}
{Writeln('The number of iterations was ', iterations : 3);}

  {Renomalise the results}
    varEpsilon := 1.0;
    for i := 1 to q do
      begin {i}
        mu[i] := mu[i] / mu[0];
        varEpsilon := varEpsilon + mu[i] * mu[i];
      end; {i}
    mu[0] := 1;
    varEpsilon := gamma[0] / varEpsilon;

  end; {Minit}


{_________________________________________________}

  procedure SeasonalPoles (var s: integer;
                  var rho: real;
                  var poles: complexVector);

    var
      theta: real;

  begin
    if (s = 12) then
      begin
        theta := (2.0 * Arctan(1.0)) / 3.0;

        poles[1].re := rho * Cos(theta);
        poles[1].im := rho * Sin(theta);

        poles[5].re := -poles[1].re;
        poles[5].im := poles[1].im;

        poles[7].re := -poles[1].re;
        poles[7].im := -poles[1].im;

        poles[11].re := poles[1].re;
        poles[11].im := -poles[1].im;

        poles[2].re := rho * Cos(2.0 * theta);
        poles[2].im := rho * Sin(2.0 * theta);

        poles[4].re := -poles[2].re;
        poles[4].im := poles[2].im;

        poles[8].re := -poles[2].re;
        poles[8].im := -poles[2].im;

        poles[10].re := poles[2].re;
        poles[10].im := -poles[2].im;

        poles[3].re := 0.0;
        poles[3].im := rho;

        poles[6].re := rho;
        poles[6].im := 0.0;

        poles[9].re := 0.0;
        poles[9].im := -rho;
      end;

    if (s = 4) then
      begin
        poles[1].re := 0.0;
        poles[1].im := rho;

        poles[2].re := -rho;
        poles[2].im := 0.0;

        poles[3].re := 0.0;
        poles[3].im := -rho;
      end;

  end; {SeasonalPoles}

{_________________________________________________}

  procedure SeasonalZeros (var s: integer;
                  var zeros: complexVector);

    var
      theta: real;

  begin
    if s = 12 then
      begin
        theta := (2.0 * Arctan(1.0)) / 3.0; {30 degrees}

        zeros[1].re := Cos(theta);
        zeros[1].im := Sin(theta);

        zeros[5].re := -zeros[1].re;
        zeros[5].im := zeros[1].im;

        zeros[7].re := -zeros[1].re;
        zeros[7].im := -zeros[1].im;

        zeros[11].re := zeros[1].re;
        zeros[11].im := -zeros[1].im;

        zeros[2].re := Cos(2.0 * theta);
        zeros[2].im := Sin(2.0 * theta);

        zeros[4].re := -zeros[2].re;
        zeros[4].im := zeros[2].im;

        zeros[8].re := -zeros[2].re;
        zeros[8].im := -zeros[2].im;

        zeros[10].re := zeros[2].re;
        zeros[10].im := -zeros[2].im;

        zeros[3].re := 0.0;
        zeros[3].im := 1.0;

        zeros[6].re := -1.0;
        zeros[6].im := 0.0;

        zeros[9].re := 0.0;
        zeros[9].im := -1.0;
      end;

    if (s = 4) then
      begin
        zeros[1].re := 0.0;
        zeros[1].im := 1.0;

        zeros[2].re := -1.0;
        zeros[2].im := 0.0;

        zeros[3].re := 0.0;
        zeros[3].im := -1.0;
      end;

  end; {SeasonalZeros}

{_________________________________________________}

  procedure SprimeY (var y: longVector;
                  s, Tcap: integer);

{THIS IS NOW REDUNDANT}

{This is a subroutine of <ApplySeasonalFilter>}
{is this seasonal cumulation? }

    var
      t, j: integer;
  begin
    for t := 0 to Tcap - s do
      for j := 1 to s - 1 do
        y[t] := y[t] + y[t + j];
  end; {SprimeY}

{_________________________________________________}

  procedure Sy2 (q, n: integer;
                  var alpha: seasVector;
                  var y: longVector);

{We need to introduce Sigmavec. What sort of strucyure does it have?}
    var
      t, j, r, s: integer;
      store: real;

  begin
    for t := n downto 0 do
      begin
        r := Max(0, q - t);
        s := Min(q, n - t);
        store := 0.0;
        for j := r to s do
          store := store + alpha[j] * y[t - q + j];
        y[t] := store;
      end;
  end; {Sy2}

{_________________________________________________}

  procedure SprimeY2 (q, n: integer;
                  var alpha: seasVector;
                  var y: longVector);

{We need to introduce Sigmavec. What sort of structure does it have?}

    var
      t, j: integer;
      store: real;

  begin
    for t := 0 to n - q do
      begin {t}
        store := 0.0;
        for j := 0 to q do
          store := store + alpha[q - j] * y[t + j];
        y[t] := store;
      end; {t}
  end; {SPrimeY2}

{_________________________________________________}

  procedure ApplySingleSeasonalFilter (var s, Tcap: integer;
                  var phi, sigmaVec: seasVector;
                  var seasonVec, adjustedVec: longVector);

{To be called by <ApplyTheFilter(s)> in ORGANISE}

{adjustedVec is a copy of residueVec--the residue from a polynonial detrending.}
{It will eventually contain the seasonaly-adjusted data. SeasonalVec, which is }
{also a copy of residueVec, will eventually contain the seasonal fluctuations.}

    var
      t: integer;

  begin
    Sprimey2(s - 1, Tcap - 1, sigmaVec, adjustedVec);
    Toeplitz(phi, adjustedVec, Tcap - s, s - 1);
    Sy2(s - 1, Tcap - 1, sigmaVec, adjustedVec);

    for t := 0 to Tcap - 1 do
      seasonVec[t] := seasonVec[t] - adjustedVec[t];

    weHaveSeasonalComponent := true;
    weHaveAdjustedResiduals := true;
    weHaveTrendCycleComponent := false;
    weHaveCycleComponent := false;

  end;{ApplySingleSeasonalFilter}

{_________________________________________________}

  procedure ApplySmoothingFilter (var s, q, Tcap: integer;
                  var phi, phi2, sigmaVec, muVec: seasVector;
                  var seasonVec, adjustedVec: longVector);
    var
      t: integer;

  begin
    Sprimey2(s - 1, Tcap - 1, sigmaVec, adjustedVec);
    Toeplitz(phi, adjustedVec, Tcap - s, s - 1);
    Sy2(s - 1, Tcap - 1, sigmaVec, adjustedVec);

    Sprimey2(q, Tcap - 1, muVec, adjustedVec);
    Sy2(q, Tcap - 1, muVec, adjustedVec);

    for t := 0 to Tcap - 1 do
      seasonVec[t] := seasonVec[t] - adjustedVec[t];

    weHaveSeasonalComponent := false;
    weHaveAdjustedResiduals := false;
    weHaveTrendCycleComponent := false;
    weHaveSeasonallyAdjustedData := false;
    weHaveCycleComponent := true;

  end; {ApplySmoothingFilter}

{_________________________________________________}

  procedure ApplyDoubleSeasonalFilter (var s, Tcap: integer;
                  var phi1, phi2, sigmaVec1, sigmaVec2: seasVector;
                  var seasonVec, adjustedVec: longVector);

    var
      t: integer;

  begin

    Sprimey2(s - 2, Tcap - 1, sigmaVec1, adjustedVec);
    Toeplitz(phi1, adjustedVec, Tcap - s + 1, s - 2);
    Sy2(s - 2, Tcap - 1, sigmaVec1, adjustedVec);

    Sprimey2(s, Tcap - 1, sigmaVec2, adjustedVec);
    Toeplitz(phi2, adjustedVec, Tcap - s - 1, s);
    Sy2(s, Tcap - 1, sigmaVec2, adjustedVec);

    for t := 0 to Tcap - 1 do
      seasonVec[t] := seasonVec[t] - adjustedVec[t];

    weHaveSeasonalComponent := true;
    weHaveSeasonallyAdjustedData := false;
    weHaveAdjustedResiduals := true;
    weHaveTrendComponent := false;
    weHaveCycleComponent := false;

  end;{ApplyDoubleSeasonalFilter}

{_________________________________________________}

  procedure ApplyTripleSeasonalFilter (var s, Tcap: integer;
                  var phi, phi1, phi2, sigmaVec, sigmaVec1, sigmaVec2: seasVector;
                  var seasonVec, adjustedVec: longVector);

    var
      t: integer;

  begin

    Sprimey2(s - 1, Tcap - 1, sigmaVec, adjustedVec);
    Toeplitz(phi, adjustedVec, Tcap - s, s - 1);
    Sy2(s - 1, Tcap - 1, sigmaVec, adjustedVec);

    Sprimey2(s - 2, Tcap - 1, sigmaVec1, adjustedVec);
    Toeplitz(phi1, adjustedVec, Tcap - s + 1, s - 2);
    Sy2(s - 2, Tcap - 1, sigmaVec1, adjustedVec);

    Sprimey2(s, Tcap - 1, sigmaVec2, adjustedVec);
    Toeplitz(phi2, adjustedVec, Tcap - 1 - s, s);
    Sy2(s, Tcap - 1, sigmaVec2, adjustedVec);

    for t := 0 to Tcap - 1 do
      seasonVec[t] := seasonVec[t] - adjustedVec[t];

    weHaveSeasonalComponent := true;
    weHaveSeasonallyAdjustedData := false;
    weHaveAdjustedResiduals := true;
    weHaveTrendComponent := false;
    weHaveCycleComponent := false;

  end;{ApplyTripleSeasonalFilter}

  {_________________________________________________}

  procedure TrendPlusSeason (var polyParas: polyParameters;
                  var dataBOX, adjustedBOx: rBOX;
                  var dataVector, adjustedVec: longVector;
                  var SCREENBOX, POSTBOX: IBOX);

    var
      z, polyt: real;
      t: integer;
      addPoly: boolean;
      legend: string;

  begin

    if weHaveCycleComponent then
      begin
        Writeln;
        Writeln('    THE TREND-CYCLE COMPONENT');
        Writeln;
        Writeln('    By  adding the cycle component to the polynomial function that has');
        Writeln('    been interpolated through the data, you will create a trend-cycle');
        Writeln('    trajectory. This will overwrite the the cyclical component.');
        Writeln;
        YesorNo('    Do you wish to overwrite cyclical component?', addPoly);
        if addPoly then
          begin
            weHaveCycleComponent := false;
            weHaveTrendCycleComponent := true;
            weHaveSeasonallyAdjustedData := false;
            legend := TrendAndCycle(dataBOX, dataName);
          end;
      end;

    if weHaveAdjustedResiduals then
      begin
        Writeln;
        Writeln('    SEASONALLY ADJUSTED TRENDED DATA');
        Writeln;
        Writeln('    By adding the seasonally adjusted version of the residual sequence');
        Writeln('    to the polynomial function that has  been interpolated through the');
        Writeln('    the data, you  will create the seasonally-adjusted data  sequence.');
        Writeln;
        {Writeln('    By creating the seasonally-adjusted trended data you will overwrite');}
        {Writeln('    the seasonally-adjusted residual sequence.');}
        Writeln;
        Writeln('                                                              <RETURN>');
        Readln;
        addPoly := true;
        {YesorNo('    Do you wish to overwrite the seasonally-adjusted residual sequence?', addPoly);}
        if addPoly then
          begin
            weHaveAdjustedResiduals := false;
            weHaveSeasonallyAdjustedData := true;
            weHaveTrendCycleComponent := false;
            legend := DeSeasonalisedOverlay(dataBOX, dataName);
          end;
      end;

    if addPoly then
      weHaveAdjustedResiduals := false;

    if addPoly then
      begin {add the polynomial}

        for t := 0 to dataBOX.NOPUNCT - 1 do
          begin {t}
            z := t;
            polyt := PolyOrdinate(z, polyParas.alpha, polyParas.gamma, polyParas.delta, polyParas.polyDegree);
            adjustedVec[t] := adjustedVec[t] + polyt;{trendVec contains the trendcycles, cycleVec contains the data}
          end; {t}

        ScreenOverplot(dataBOX, SCREENBOX, dataVector, adjustedVec, legend);

        if (PSPict = true) then
          WritePostOverFile(dataBOX, POSTBOX, dataVector, adjustedVec, legend);

      end; {add the polynomial}

  end; {TrendPlusSeason}

{________________________________________________}


end. {Unit 9 of SEASIDE.PAS}

